﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading;

namespace ThreadSafeEvents {

	class Control1 {
		public event EventHandler Click;
	}

	class Control2_inCSharp3 {
		private EventHandler __Click;

		public event EventHandler Click {
			add { lock (this) { __Click = __Click + value; } }
			remove { lock (this) { __Click = __Click - value; } }
		}
	}

	class Control2_inCSharp4 {
		private EventHandler __Click;

		public event EventHandler Click {
			add { 
				EventHandler oldHandler, newHandler;
				do {
					oldHandler = __Click;
					newHandler = oldHandler + value;
				} while (Interlocked.CompareExchange(ref __Click, newHandler, oldHandler) != oldHandler);
			}
			remove { /* similar thread-safe and lock-free __Click = __Click - value; */ }
		}
	}

	class Control3 {
		public event EventHandler Click;

		protected void OnClick(EventArgs e) {
			if (Click != null) {
				Click(this, e);
			}
		}
	}

	class Control4 {
		private EventHandler click;
		private object clickRaiseLock = new object();

		public event EventHandler Click {
			add { lock (clickRaiseLock) { click = click + value; } }
			remove { lock (clickRaiseLock) { click = click - value; } }
		}

		protected void OnClick(EventArgs e) {
			lock (clickRaiseLock) {
				if (click != null) {
					click(this, e);
				}
			}
		}
	}

	class Control5 {
		public event EventHandler Click = (sender, eventArgs) => { };

		protected void OnClick(EventArgs e) {
			Click(this, e);
		}
	}

	class Control6 {
		public event EventHandler Click;

		protected void OnClick(EventArgs e) {
			EventHandler handler = Click;
			if (handler != null) {
				handler(this, e);
			}
		}
	}

	// by James Michael Hare

	static class EventExtensions {
		public static void Raise(this EventHandler handler, object sender, EventArgs e) {
			if (handler != null) {
				handler(sender, e);
			}
		}

		public static void Raise<T>(this EventHandler<T> handler, object sender, T e) where T : EventArgs {
			if (handler != null) {
				handler(sender, e);
			}
		}
	}

	class Control7 {
		public event EventHandler Click;

		protected void OnClick(EventArgs e) {
			Click.Raise(this, e);
		}
	}

	class Control8 { // Solution in C# 6
		public event EventHandler Click;

		protected void OnClick(EventArgs e) {
			Click?.Invoke(this, e);
			// The new way is thread-safe because the compiler generates code to evaluate Click one time only, keeping the result in a temporary variable. 
		}
	}

	class Program {
		static void Main(string[] args) {
		}
	}
}
